# PersonalFME - Gateway linking analog radio selcalls to internet communication services
# Copyright(C) 2010-2017 Ralf Rettig (www.personalfme.de)
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.If not, see <http://www.gnu.org/licenses/>

# Creates a target for generating a ZIP-file containing the source code of PersonalFME
# Parameters:
# ZIP_FILE_NAME					- Name of the ZIP-file
# EXCLUDED_FILES				- List of excluded files
# EXCLUDED_DIRS					- List of excluded directories
# ADDITIONAL_FILES				- List of additionally included files
#
# * all files need to be provided relative to the project source root directory, all file names are case-sensitive
# on Linux
# * 7z is required on Windows for creating the ZIP-file

function( CreateSourceZIPTarget
	ZIP_FILE_NAME
	EXCLUDED_FILES
	EXCLUDED_DIRS
	ADDITIONAL_FILES
)
	find_package( Subversion )
	Subversion_WC_INFO( ${PROJECT_SOURCE_DIR} Project )
	
	set( TEMP_DIR temp )
	
	# create lists with the full names of the excluded files and directories
	foreach( file_name ${EXCLUDED_FILES} )
		list( APPEND EXCLUDED_FILES_LIST ${CMAKE_BINARY_DIR}/${TEMP_DIR}/${file_name} )
	endforeach()
	foreach( dir_name ${EXCLUDED_DIRS} )
		list( APPEND EXCLUDED_DIRS_LIST ${CMAKE_BINARY_DIR}/${TEMP_DIR}/${dir_name} )
	endforeach()	

	if ( ${Option_CREATE_DOCUMENTATION} )
		add_custom_target( SourceZIPFile DEPENDS Documentation )
	else()
		add_custom_target( SourceZIPFile )
	endif()
	set_target_properties( SourceZIPFile PROPERTIES EXCLUDE_FROM_ALL ON EXCLUDE_FROM_DEFAULT_BUILD ON )

	# clear any existing ZIP-file and temporary directory	
	add_custom_command( TARGET SourceZIPFile 
						COMMAND ${CMAKE_COMMAND} -E remove ${ZIP_FILE_NAME} )
	add_custom_command( TARGET SourceZIPFile 
						COMMAND ${CMAKE_COMMAND} -E remove_directory ${CMAKE_BINARY_DIR}/${TEMP_DIR} )
	
	# export a clean copy of the current revision from the SVN-server
	add_custom_command( TARGET SourceZIPFile
						COMMAND ${Subversion_SVN_EXECUTABLE} export ${Project_WC_URL}@${Project_WC_REVISION} ${CMAKE_BINARY_DIR}/${TEMP_DIR} --ignore-externals )
		
	# remove the excluded files and directories
	add_custom_command( TARGET SourceZIPFile 
						COMMAND ${CMAKE_COMMAND} -E remove ${EXCLUDED_FILES_LIST} )
	foreach( dir_path ${EXCLUDED_DIRS_LIST} )
		add_custom_command( TARGET SourceZIPFile 
							COMMAND ${CMAKE_COMMAND} -E remove_directory ${dir_path} )
	endforeach()
	
	# add the additional files
	foreach( added_file ${ADDITIONAL_FILES} )
		add_custom_command( TARGET SourceZIPFile 
							COMMAND ${CMAKE_COMMAND} -E copy ${PROJECT_SOURCE_DIR}/${added_file} ${CMAKE_BINARY_DIR}/${TEMP_DIR}/${added_file} )
	endforeach()
		
	# create the ZIP-file		
	if( ${CMAKE_SYSTEM_NAME} MATCHES "Windows" )
		if ( ${Option_CREATE_DOCUMENTATION} )
			message( "Info: Visual Studio will not create or update the documentation target automatically when running the source code ZIP-file target!")
		endif()
	endif()
	
	message( "Info: 7-Zip is required to create the source code ZIP-file artifact!")
	add_custom_command( TARGET SourceZIPFile 
						COMMAND 7z a ${ZIP_FILE_NAME} ${CMAKE_BINARY_DIR}/${TEMP_DIR}/* )
		
	# clean up the temp directory
	add_custom_command( TARGET SourceZIPFile 
						COMMAND ${CMAKE_COMMAND} -E remove_directory ${CMAKE_BINARY_DIR}/${TEMP_DIR} )
endfunction()
